#include "WPILib.h"
#include "Drive.h"
#include "IOMap.h"

const double kAnnoyanceSec = 0.5;

const float kEncPerRotation = 256;
const float kInchPerRotation = 12.57; //4in wheel
const float kStraightRange = 40;
const double kStraightChange = .025;
const float kRampDistance = 250; //in encoder ticks, about 6 in
const float kRampIncrease = .1;
const float kDriveAutoAngleTolerance = 13;
const double MOTOR_RAMP_PER_CYCLE = .2;
const double kStraightGyroRange = .4;
const int kLostIt = 5;


Drive::Drive():
	leftEncoder(0,1),
	rightEncoder(2,3),
	gyro(1),
	_leftSpeed(0),
	_rightSpeed(0),
	_autoDistance(0),
	drive(kTeleop),
	_distance(0),
	_angle(0),
	_autoTurnFinished(true),
	_autoStraightFinished(0),
	_startGyro(0),
	_autoSpeed(0),
	_leftFinalSpeed(0),
	_rightFinalSpeed(0),
	_rampTimes(1),
	_wantSpeed(0),
	_mathAngle(0),
	_rightEncBroke(false),
	_leftEncBroke(false),
	_gyroBroke(false),
	_gyroCalibrated(false),
	_currentDriveLeft(0),
	_currentDriveRight(0),
	_horizontalOffset(0),
	_distanceBetween(0),
	_gyroAnnoyanceTimer(),
	_annoyanceVal(false),
	_startVisionTracker(false),
	_visionCounter(0),
	_visionResult(0),
	_lastVisionCounter(0),
	_visionCountChanged(false),
	_lastStartVision(true),
	_haveSeen(false),
	timer(),
	_leftMotorVision(0),
	_rightMotorVision(0),
	_visionOffsetTolerance(12),
	_lostItCounter(0)


	{
		leftMotor = new Spark(PWM_OUT_DRIVE_LEFT);
		leftMini = new Spark(PWM_OUT_MINI_LEFT);
		rightMotor = new Spark(PWM_OUT_DRIVE_RIGHT);
		rightMini = new Spark(PWM_OUT_MINI_RIGHT);
		_gyroAnnoyanceTimer.Start();
		timer.Start();
	}


	void Drive::Move_Speed(double leftSpeed, double rightSpeed){
		_leftFinalSpeed = leftSpeed;
		_rightFinalSpeed = rightSpeed;
		drive = kTeleop;
	}

	void Drive::Auto_Straight(double speed, double distance){
		_rampTimes = 1;
		_autoStraightFinished = false;
		_leftSpeed = 0;
		_rightSpeed = 0;
		_wantSpeed = speed;
		rightEncoder.Reset();
		leftEncoder.Reset();
		gyro.Reset();
		drive = kAutoStraight;
		_distance = (distance/kInchPerRotation)*kEncPerRotation;
	}

	void Drive::Auto_Pivot_Turn(float angle){
		_autoTurnFinished = false;
		gyro.Reset();
		_angle = angle;
		_autoSpeed = .35;
		drive = kAutoTurn;
	}

	void Drive::Auto_Vision_Straight(){
		timer.Reset();
		_lastVisionCounter = _visionCounter;	// sync up with the dashboard
		_startVisionTracker = true;				// tell dashboard to start
		_autoStraightFinished = false;
		_wantSpeed = .31;
		_visionOffsetTolerance = 10;
		_lostItCounter = 0;
		drive = kAutoVisionStraight;
	}

	void Drive::Auto_Vision_Turn(bool left){
		timer.Reset();
		_lastVisionCounter = _visionCounter;	// sync up with the dashboard
		_startVisionTracker = true;				// tell dashboard to start
		_autoTurnFinished = false;
		_wantSpeed = .275;
		_visionOffsetTolerance = 8;
		_lostItCounter = 0;
		drive = kAutoVisionTurn;
	}

	bool Drive::Is_Auto_Straight_Finished(){
		if(_autoStraightFinished){
			return true;
		}
		else{
			return false;
		}
	}
	bool Drive::Is_Auto_Turn_Finished(){
		if(_autoTurnFinished){
			return true;
		}
		else{
			return false;
		}
	}
	void Drive::Process(){
		switch (drive){
			case kTeleop:

				// Ramping for the left motors
				if (_currentDriveLeft > _leftFinalSpeed)
				{
					_currentDriveLeft -= MOTOR_RAMP_PER_CYCLE;
					if (_currentDriveLeft < _leftFinalSpeed)
					{
						_currentDriveLeft = _leftFinalSpeed;
					}
				}else if (_currentDriveLeft < _leftFinalSpeed)
				{
					_currentDriveLeft += MOTOR_RAMP_PER_CYCLE;
					if (_currentDriveLeft > _leftFinalSpeed)
					{
						_currentDriveLeft = _leftFinalSpeed;
					}
				}else if (_currentDriveLeft == _leftFinalSpeed)
				{
				}

				// Ramping for the Right motors
				if (_currentDriveRight > _rightFinalSpeed)
				{//If the speed we are going at now is more than the speed we wish to go at
					_currentDriveRight -= MOTOR_RAMP_PER_CYCLE; //reduce the speed we are going at
					if (_currentDriveRight < _rightFinalSpeed) //If we have overshot the goal
					{
						_currentDriveRight = _rightFinalSpeed; //set the current speed to the goal
					}
				}else if (_currentDriveRight < _rightFinalSpeed)//repeat the same, except we are now greater rather than lower
				{
					_currentDriveRight += MOTOR_RAMP_PER_CYCLE;
					if (_currentDriveRight > _rightFinalSpeed)
					{
						_currentDriveRight = _rightFinalSpeed;
					}
				}else if (_currentDriveRight == _rightFinalSpeed)
				{
				}

				break;

			case kAutoStraight:
				_leftFinalSpeed = 0;
				_rightFinalSpeed = 0;
				if(!_leftEncBroke && !_rightEncBroke){
					_autoDistance = (Get_Left_Encoder() + Get_Right_Encoder())/2;	//current distance, average of the two encoder values
				}
				else if(_leftEncBroke && !_rightEncBroke){
					_autoDistance = Get_Right_Encoder();
				}
				else if(_rightEncBroke && !_leftEncBroke){
					_autoDistance = Get_Left_Encoder();
				}
				if(_autoStraightFinished == false && (!_leftEncBroke || !_rightEncBroke)){
					if((_distance - (_rampTimes*kRampDistance) < abs(_autoDistance)) && (abs(_leftSpeed) -kRampIncrease) > 0){
						if(_wantSpeed > 0){
							_leftSpeed = _leftSpeed-kRampIncrease;
							_rightSpeed = _rightSpeed - kRampIncrease;
							_rampTimes--;
						}
						else if(_wantSpeed < 0){
							_leftSpeed = _leftSpeed + kRampIncrease;
							_rightSpeed = _rightSpeed + kRampIncrease;
							_rampTimes--;
						}
						else{
							_autoStraightFinished = true;
							_leftSpeed = 0;
							_rightSpeed = 0;
						}
					}
					else if((_rampTimes*kRampIncrease) + kRampIncrease < abs(_wantSpeed)){
						if(abs(_autoDistance) < _rampTimes*kRampDistance){
							if(_wantSpeed > 0){
								_leftSpeed = _leftSpeed + kRampIncrease;
								_rightSpeed = _rightSpeed + kRampIncrease;
								_rampTimes++;
							}
							else if(_wantSpeed < 0){
								_leftSpeed = _leftSpeed - kRampIncrease;
								_rightSpeed = _rightSpeed - kRampIncrease;
								_rampTimes++;
							}
						}
					}
					else{
						_leftSpeed = _wantSpeed;
						_rightSpeed = _wantSpeed;
					}
				}
				_leftSpeed = _wantSpeed;
				_rightSpeed = _wantSpeed;
				if(!_gyroBroke && !_autoStraightFinished){
					if(gyro.GetAngle() < kStraightGyroRange && _wantSpeed > 0){
						_leftSpeed = _leftSpeed + kStraightChange;
						_rightSpeed = _rightSpeed - kStraightChange;
					}
					else if(gyro.GetAngle() > kStraightGyroRange && _wantSpeed > 0){
						_leftSpeed = _leftSpeed - kStraightChange;
						_rightSpeed = _rightSpeed + kStraightChange;
					}
					else if(gyro.GetAngle() < kStraightGyroRange && _wantSpeed < 0){
						_leftSpeed = _leftSpeed - kStraightChange;
						_rightSpeed = _rightSpeed + kStraightChange;
					}
					else if(gyro.GetAngle() > kStraightGyroRange && _wantSpeed < 0){
						_leftSpeed = _leftSpeed + kStraightChange;
						_rightSpeed = _rightSpeed - kStraightChange;
					}
					_leftFinalSpeed = _leftSpeed;
					_rightFinalSpeed = _rightSpeed;
				}
				if(_leftEncBroke && _rightEncBroke){
					_leftFinalSpeed = 0;
					_rightFinalSpeed = 0;
				}

				if((_autoDistance < _distance && _wantSpeed > 0) || (_autoDistance > -_distance && _wantSpeed < 0)){	//if our current position is less than we want...
					//keep going dont change numbers
				}
				else{
					_leftFinalSpeed = 0;
					_rightFinalSpeed = 0;
					_autoStraightFinished = true;
				}
				break;

			case kAutoTurn:
				_leftFinalSpeed = 0;
				_rightFinalSpeed = 0;
				if(_autoTurnFinished == false){
							if(_angle < 0){
								_leftFinalSpeed = -_autoSpeed;
								_rightFinalSpeed = _autoSpeed; //-.6
							}
							else if(_angle > 0){

								_leftFinalSpeed = _autoSpeed;
								_rightFinalSpeed = -_autoSpeed;
							}
					if(gyro.GetAngle() > _angle - kDriveAutoAngleTolerance && _angle > 0){
								_leftFinalSpeed = 0;
								_rightFinalSpeed = 0;
								_autoTurnFinished = true;
								gyro.Reset();
						}
							else if(gyro.GetAngle() < _angle + kDriveAutoAngleTolerance && _angle < 0){
								_leftFinalSpeed = 0;
								_rightFinalSpeed = 0;
								_autoTurnFinished = true;
								gyro.Reset();
							}

				}
				break;
			case kAutoVisionTurn:
				_leftFinalSpeed = 0;
				_rightFinalSpeed = 0;
				if(!_autoTurnFinished){
					printf("Attempting Vision Turn %d %d %d %d\n", _autoTurnFinished, _visionCountChanged, _visionResult, _horizontalOffset);
					if (!CheckTarget(_wantSpeed, -1)){
						_leftFinalSpeed = _leftMotorVision;
						_rightFinalSpeed = _rightMotorVision;
					}
					else {
						printf("Done\n");
						_autoTurnFinished = true;
						_startVisionTracker = false;
					}
				}
				break;
			case kAutoVisionStraight:
				_leftFinalSpeed = 0;
				_rightFinalSpeed = 0;
				if(!_autoStraightFinished){
					if (!CheckTarget(.075, 50)){
						_leftFinalSpeed = _wantSpeed + _leftMotorVision;
						_rightFinalSpeed = _wantSpeed + _rightMotorVision;
					}
					else {
						printf("Done\n");
						_autoStraightFinished = true;
						_startVisionTracker = false;
						_leftFinalSpeed = 0;
						_rightFinalSpeed = 0;
					}
				}
				break;
		}
		printf("%f %f", _leftFinalSpeed, _rightFinalSpeed);
		Set_Left_Motors(-_leftFinalSpeed);	//sets each motor to the final speed value
		Set_Right_Motors(_rightFinalSpeed);
	}

	int Drive::Get_Right_Encoder(){
		return(-rightEncoder.Get());
	}

	int Drive::Get_Left_Encoder(){
		return(leftEncoder.Get());
	}

	void Drive::Left_Encoder_Set_Broken(bool broken){
		_leftEncBroke = broken;
	}

	void Drive::Right_Encoder_Set_Broken(bool broken){
		_rightEncBroke = broken;
	}

	void Drive::Gyro_Set_Broken(bool broken) {
		_gyroBroke = broken;
	}

	void Drive::Calibrate_Gyro(){
		if(!_gyroCalibrated){
			gyro.Calibrate();
			_gyroAnnoyanceTimer.Stop();
			SmartDashboard::PutNumber("thunderdashboard_gyro",0);
			_gyroCalibrated = true;
		}
	}

	void Drive::Reset(){
		_leftSpeed = 0;
		_rightSpeed = 0;
		_distance = 0;
		_angle =0;
		_autoTurnFinished = true;
		_autoStraightFinished = true;
		_startGyro = 0;
		Calibrate_Gyro();
		gyro.Reset();
		drive = kTeleop;
		Set_Left_Motors(0);	//sets each motor to the final speed value
		Set_Right_Motors(0);
		rightEncoder.Reset();
		leftEncoder.Reset();
		_currentDriveLeft = 0;
		_currentDriveRight = 0;
		_leftFinalSpeed = 0;
		_rightFinalSpeed = 0;

	}

	void Drive::Debug(Feedback *feedback){
		feedback->send_Debug_Double("drive", "leftEncoder", Get_Left_Encoder());
		feedback->send_Debug_Double("drive", "rightEncoder", Get_Right_Encoder());
		feedback->send_Debug_Double("drive", "rightSpeed", _rightFinalSpeed);
		feedback->send_Debug_Double("drive", "leftSpeed", _leftFinalSpeed);
		feedback->send_Debug_Double("drive", "gyro_angle", gyro.GetAngle());
		feedback->send_Debug_Double("drive", "angle_wanted", _angle);
		feedback->send_Debug_String("drive", "is_turn_finished", _autoTurnFinished ? "true":"false");
		feedback->send_Debug_String("drive", "is_Straight_finished", _autoStraightFinished ? "true":"false");
		feedback->send_Debug_String("drive", "left_encoder_broken?", _leftEncBroke ? "true":"false");
		feedback->send_Debug_String("drive", "right_encoder_broken?", _rightEncBroke ? "true":"false");
		feedback->send_Debug_String("drive", "is_gyro_calibrated?", _gyroCalibrated ? "true":"false");
		if(_startVisionTracker != _lastStartVision){
			feedback->send_Debug_String("vision", "Start", _startVisionTracker ? "true":"false");
			_lastStartVision = _startVisionTracker;
		}

		if(!_gyroCalibrated){
			if (_gyroBroke) {
				SmartDashboard::PutNumber("thunderdashboard_gyro", 1);
			} else if (_gyroAnnoyanceTimer.Get() > kAnnoyanceSec) {
				SmartDashboard::PutNumber("thunderdashboard_gyro", _annoyanceVal ? 1 : 0);
				_annoyanceVal = !_annoyanceVal;
				_gyroAnnoyanceTimer.Reset();
			}
		}

		_horizontalOffset = (int)SmartDashboard::GetNumber("vision_HorizontalOffset", 0);
		_distanceBetween = (int)SmartDashboard::GetNumber("vision_DistanceBetween", 0);
		_visionResult = (int)SmartDashboard::GetNumber("vision_Result", 0);
		_visionCounter = (int)SmartDashboard::GetNumber("vision_Counter", 0);
	}

	void Drive::Set_Left_Motors(double Power){
		leftMotor->Set(Power);
		leftMini->Set(Power);
	}

	void Drive::Set_Right_Motors(double Power){
		rightMotor->Set(Power);
		rightMini->Set(Power);
	}

	bool Drive::Did_Auto_Die(){
		return(_lostItCounter > kLostIt);
	}

	bool Drive::CheckTarget(double speed, int distance){
		// assume no turning should happen
		_leftMotorVision = 0;
		_rightMotorVision = 0;

		// see if there is a new value to read or not
		if (_visionCounter == _lastVisionCounter){
			printf("No new info\n");
			_lostItCounter ++;
			if(_lostItCounter > kLostIt){
				return(true);
			}
			return (false);				// have no idea if we are on target or not
		}
		_lastVisionCounter = _visionCounter;

		if (_visionResult != 1){
			// target wasn't valid
			printf("invalid target\n");
				_lostItCounter ++;
				if(_lostItCounter > kLostIt){
					return(true);
				}
			return (false);				// have no idea if we are on target or not
		}
		_lostItCounter = 0;

		// if we need to check the distance, then do it now before we
		if (distance != -1) {
			if(_distanceBetween > distance){
				_leftMotorVision = 0;
				_rightMotorVision = 0;
				return true;
			}
		}

		if (_horizontalOffset > _visionOffsetTolerance){
			// target in right side of image, so turn right
			printf("> kVisionOffsetTolerance %d\n", _horizontalOffset);
			_leftMotorVision = speed;
			_rightMotorVision = -speed;
			return (false);
		}

		if (_horizontalOffset < -_visionOffsetTolerance){
			// target in left side of image, so turn left
			printf("< -kVisionOffsetTolerance %d\n", _horizontalOffset);
			_leftMotorVision = -speed;
			_rightMotorVision = speed;
			return (false);
		}

		printf("on target\n");
		return (distance == -1);
	}

