/*----------------------------------------------------------------------------*/
/* Copyright (c) FIRST 2016. All Rights Reserved.                             */
/* Open Source Software - may be modified and shared by FRC teams. The code   */
/* must be accompanied by the FIRST BSD license file in the root directory of */
/* the project.                                                               */
/*----------------------------------------------------------------------------*/

#include "ThunderCameraServer.h"
#include "WPIErrors.h"
#include "Utility.h"

#include <iostream>
#include <chrono>
#include <cstring>
#include <sys/socket.h>
#include <unistd.h>
#include <netdb.h>

constexpr uint8_t ThunderCameraServer::kMagicNumber[];

ThunderCameraServer* ThunderCameraServer::GetInstance() {
  static ThunderCameraServer instance;
  return &instance;
}

ThunderCameraServer::ThunderCameraServer()
    : m_cameras(),
	  m_curCamIndex(kInvalidCamIndex),
      m_serverThread(&ThunderCameraServer::Serve, this),
      m_captureThread(),
      m_imageMutex(),
      m_newImageVariable(),
      m_dataPool(3),
      m_quality(50),
      m_autoCaptureStarted(false),
	  m_autoCaptureStop(false),
      m_hwClient(true),
      m_imageData(nullptr, 0, 0, false) {
  for (int i = 0; i < 3; i++) m_dataPool.push_back(new uint8_t[kMaxImageSize]);
}

void ThunderCameraServer::FreeImageData(
    std::tuple<uint8_t*, unsigned int, unsigned int, bool> imageData) {
  if (std::get<3>(imageData))
    imaqDispose(std::get<0>(imageData));
  else if (std::get<0>(imageData) != nullptr) {
    std::lock_guard<priority_recursive_mutex> lock(m_imageMutex);
    m_dataPool.push_back(std::get<0>(imageData));
  }
}

void ThunderCameraServer::SetImageData(uint8_t* data, unsigned int size,
                                unsigned int start, bool imaqData) {
  std::lock_guard<priority_recursive_mutex> lock(m_imageMutex);
  FreeImageData(m_imageData);
  m_imageData = std::make_tuple(data, size, start, imaqData);
  m_newImageVariable.notify_all();
}

void ThunderCameraServer::SetImage(Image const* image) {
  unsigned int dataSize = 0;
  uint8_t* data =
      (uint8_t*)imaqFlatten(image, IMAQ_FLATTEN_IMAGE, IMAQ_COMPRESSION_JPEG,
                            10 * m_quality, &dataSize);

  // If we're using a HW camera, then find the start of the data
  bool hwClient;
  {
    // Make a local copy of the hwClient variable so that we can safely use it.
    std::lock_guard<priority_recursive_mutex> lock(m_imageMutex);
    hwClient = m_hwClient;
  }
  unsigned int start = 0;
  if (hwClient) {
    while (start < dataSize - 1) {
      if (data[start] == 0xFF && data[start + 1] == 0xD8)
        break;
      else
        start++;
    }
  }
  dataSize -= start;

  wpi_assert(dataSize > 2);
  SetImageData(data, dataSize, start, true);
}

void ThunderCameraServer::AutoCapture() {
  Image* frame = imaqCreateImage(IMAQ_IMAGE_RGB, 0);

  while (true) {
    bool hwClient;
    std::shared_ptr<nivision::USBCamera> camPtr;
    uint8_t* data = nullptr;
    {
      std::lock_guard<priority_recursive_mutex> lock(m_imageMutex);
      if (m_autoCaptureStop) {
    	  m_autoCaptureStarted = false;
    	  break;
      }

      hwClient = m_hwClient;
      if (hwClient) {
        data = m_dataPool.back();
        m_dataPool.pop_back();
      }
      camPtr = m_cameras[m_curCamIndex];
    }

    if (hwClient) {
      unsigned int size = camPtr->GetImageData(data, kMaxImageSize);
      SetImageData(data, size);
    } else {
      camPtr->GetImage(frame);
      SetImage(frame);
    }
  }
}

void ThunderCameraServer::StartAutomaticCapture(char const* cameraName) {
	std::vector<std::string> v;
	v.push_back(cameraName);
	StartAutomaticCapture(v);
}

void ThunderCameraServer::StartAutomaticCapture(std::vector<std::string> cameraNames) {
  if (cameraNames.empty())
    return;
  std::lock_guard<priority_recursive_mutex> lock(m_imageMutex);
  if (m_autoCaptureStarted) return;
  std::vector<std::shared_ptr<nivision::USBCamera> > camVec;
  for (size_t i = 0; i < cameraNames.size(); ++i) {
    std::shared_ptr<nivision::USBCamera> camera =
      std::make_shared<nivision::USBCamera>(cameraNames[i].c_str(), true);
    camera->OpenCamera();
    camVec.push_back(camera);
  }
  StartAutomaticCapture(camVec);
}

void ThunderCameraServer::StartAutomaticCapture(std::shared_ptr<nivision::USBCamera> camera) {
  std::vector<std::shared_ptr<nivision::USBCamera> > v;
  v.push_back(camera);
  StartAutomaticCapture(v);
}

void ThunderCameraServer::StartAutomaticCapture(std::vector<std::shared_ptr<nivision::USBCamera> > cameras) {
  if (cameras.empty())
    return;
  std::lock_guard<priority_recursive_mutex> lock(m_imageMutex);
  if (m_autoCaptureStarted) return;

  m_cameras = cameras;

  m_captureThread = std::thread(&ThunderCameraServer::AutoCapture, this);
  //m_captureThread.detach();
  m_autoCaptureStarted = true;

  SetCurrentCameraIndex(0);
}

void ThunderCameraServer::SetCurrentCameraIndex(int n)
{
  std::lock_guard<priority_recursive_mutex> lock(m_imageMutex);
  if (n < 0 || (size_t)n >= m_cameras.size())
	  return;
  if (m_curCamIndex != kInvalidCamIndex)
	  m_cameras[m_curCamIndex]->StopCapture();
  m_curCamIndex = n;
  m_cameras[m_curCamIndex]->StartCapture();
}


void ThunderCameraServer::StopAutomaticCapture()
{
  {
    std::lock_guard<priority_recursive_mutex> lock(m_imageMutex);
    if (!m_autoCaptureStarted) return;

    m_autoCaptureStop = true;
  }
  m_captureThread.join();
}

bool ThunderCameraServer::IsAutoCaptureStarted() {
  std::lock_guard<priority_recursive_mutex> lock(m_imageMutex);
  return m_autoCaptureStarted;
}

void ThunderCameraServer::SetSize(unsigned int size) {
  std::lock_guard<priority_recursive_mutex> lock(m_imageMutex);
  for (size_t i = 0; i < m_cameras.size(); ++i) {
	  nivision::USBCamera *cam = m_cameras[i].get();
    if (!cam) return;
    if (size == kSize160x120)
      cam->SetSize(160, 120);
    else if (size == kSize320x240)
      cam->SetSize(320, 240);
    else if (size == kSize640x480)
      cam->SetSize(640, 480);
  }
}

void ThunderCameraServer::SetQuality(unsigned int quality) {
  std::lock_guard<priority_recursive_mutex> lock(m_imageMutex);
  m_quality = quality > 100 ? 100 : quality;
}

unsigned int ThunderCameraServer::GetQuality() {
  std::lock_guard<priority_recursive_mutex> lock(m_imageMutex);
  return m_quality;
}

void ThunderCameraServer::Serve() {
  int sock = socket(AF_INET, SOCK_STREAM, 0);

  if (sock == -1) {
    wpi_setErrnoError();
    return;
  }

  int reuseAddr = 1;
  if (setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &reuseAddr,
                 sizeof(reuseAddr)) == -1)
    wpi_setErrnoError();

  sockaddr_in address, clientAddress;

  memset(&address, 0, sizeof(address));
  address.sin_family = AF_INET;
  address.sin_addr.s_addr = htonl(INADDR_ANY);
  address.sin_port = htons(kPort);

  if (bind(sock, (struct sockaddr*)&address, sizeof(address)) == -1)
    wpi_setErrnoError();

  if (listen(sock, 10) == -1) wpi_setErrnoError();

  while (true) {
    socklen_t clientAddressLen = sizeof(clientAddress);

    int conn =
        accept(sock, (struct sockaddr*)&clientAddress, &clientAddressLen);
    if (conn == -1) {
      wpi_setErrnoError();
      continue;
    }

    Request req;
    if (read(conn, &req, sizeof(req)) == -1) {
      wpi_setErrnoError();
      close(conn);
      continue;
    } else {
      req.fps = ntohl(req.fps);
      req.compression = ntohl(req.compression);
      req.size = ntohl(req.size);
    }

    // TODO: Support the SW Compression. The rest of the code below will work as
    // though this
    // check isn't here
    if (req.compression != kHardwareCompression) {
      wpi_setWPIErrorWithContext(IncompatibleState,
                                 "Choose \"USB Camera HW\" on the dashboard");
      close(conn);
      continue;
    }

    {
      // Wait for the camera to be setw
      std::unique_lock<priority_recursive_mutex> lock(m_imageMutex);
      if (m_curCamIndex == kInvalidCamIndex) {
        std::cout << "Camera not yet ready, awaiting first image" << std::endl;
        m_newImageVariable.wait(lock);
      }
      m_hwClient = req.compression == kHardwareCompression;
      if (!m_hwClient)
        SetQuality(100 - req.compression);
      else {
        for (size_t i = 0; i < m_cameras.size(); ++i)
          m_cameras[i]->SetFPS(req.fps);
      }
      SetSize(req.size);
    }

    auto period = std::chrono::microseconds(1000000) / req.fps;
    while (true) {
      auto startTime = std::chrono::steady_clock::now();
      std::tuple<uint8_t*, unsigned int, unsigned int, bool> imageData;
      {
        std::unique_lock<priority_recursive_mutex> lock(m_imageMutex);
        m_newImageVariable.wait(lock);
        imageData = m_imageData;
        m_imageData = std::make_tuple<uint8_t*>(nullptr, 0, 0, false);
      }

      unsigned int size = std::get<1>(imageData);
      unsigned int netSize = htonl(size);
      unsigned int start = std::get<2>(imageData);
      uint8_t* data = std::get<0>(imageData);

      if (data == nullptr) continue;

      if (write(conn, kMagicNumber, sizeof(kMagicNumber)) == -1) {
        wpi_setErrnoErrorWithContext(
            "[CameraServer] Error sending magic number");
        FreeImageData(imageData);
        break;
      }
      if (write(conn, &netSize, sizeof(netSize)) == -1) {
        wpi_setErrnoErrorWithContext("[CameraServer] Error sending image size");
        FreeImageData(imageData);
        break;
      }
      if (write(conn, &data[start], sizeof(uint8_t) * size) == -1) {
        wpi_setErrnoErrorWithContext("[CameraServer] Error sending image data");
        FreeImageData(imageData);
        break;
      }
      FreeImageData(imageData);
      std::this_thread::sleep_until(startTime + period);
    }
    close(conn);
  }
  close(sock);
}
